/*  hivemind -- a client for cadaver
    Copyright (C) 2000 Kari Pahula

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
*/

#define CADAVER_PORT 8932
#define BUF_SIZE (1<<13)

#include <config.h>

#include <netdb.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>
#include <getopt.h>

#ifdef HAVE_LIBPTHREAD
#include <pthread.h>
#endif

#include <iostream>
#include <vector>
#include <set>
#include <algorithm>
#include <string>

#include "turf.H"
#include "control.H"
#include "nomad.H"

class Time {
public:
  void now_time () {gettimeofday (&tv, NULL);}
  Time (const struct timeval &set) {tv = set;}
  Time () {}

  Time operator+ (const Time &) const;
  Time operator- (const Time &) const;
  Time & operator+= (const Time &);
  Time & operator-= (const Time &);
  Time & operator+= (const int &t) {tv.tv_usec+=t; normalize ();}
  Time & operator= (const int &t) {tv.tv_sec=0; tv.tv_usec=t; normalize ();}
  bool operator>= (const Time &val) const;

  operator struct timeval () const {return tv;}
  operator struct timeval * () {return &tv;}
private:
  void normalize ();
  struct timeval tv;
};

Time Time::operator+ (const Time &val) const
{
  Time ret;
  ret.tv.tv_sec = tv.tv_sec+val.tv.tv_sec;
  ret.tv.tv_usec = tv.tv_usec+val.tv.tv_usec;
  ret.normalize ();
  return ret;
}

Time Time::operator- (const Time &val) const
{
  Time ret;
  ret.tv.tv_sec = tv.tv_sec-val.tv.tv_sec;
  ret.tv.tv_usec = tv.tv_usec-val.tv.tv_usec;
  ret.normalize ();
  return ret;
}

Time & Time::operator+= (const Time &val)
{
  tv.tv_sec+=val.tv.tv_sec;
  tv.tv_usec+=val.tv.tv_usec;
  normalize ();
  return *this;
}

Time & Time::operator-= (const Time &val)
{
  tv.tv_sec-=val.tv.tv_sec;
  tv.tv_usec-=val.tv.tv_usec;
  normalize ();
  return *this;
}

bool Time::operator>= (const Time &val) const
{
  return tv.tv_sec > val.tv.tv_sec || (tv.tv_sec == val.tv.tv_sec
					&& tv.tv_usec >= val.tv.tv_usec);
}

void Time::normalize ()
{
  while (tv.tv_usec >= 1000000)
    {
      tv.tv_usec -= 1000000;
      tv.tv_sec++;
    }
  while (tv.tv_usec < 0)
    {
      tv.tv_usec += 1000000;
      tv.tv_sec--;
    }
}

int Cyborg::pseudofloat (const char **msg)
{
  const char *src;
  char *dst, buf[14];
  dst = &buf[13];
  buf[13] = '\0';

  while (**msg == ' ') (*msg)++;
  while (**msg != '.') (*msg)++;
  (*msg)+= 2;
  src = (*msg)-1;

  while (*src != ' ')
    {
      if (*src != '.')
	*(--dst) = *src;
      src--;
    }
  return strtol (dst, NULL, 10);
}

Cyborg::Cyborg (const char *&msg)
{
  while (*msg == ' ') msg++;
  gender = *msg == 'T' ? true : false;
  state = static_cast <enum state>
    (strtol (msg+2, const_cast <char **> (&msg), 10));
  item = static_cast <enum item>
    (strtol (msg+1, const_cast <char **> (&msg), 10));
  targetid = strtol (msg+1, const_cast <char **> (&msg), 10);
  id = strtol (msg+1, const_cast <char **> (&msg), 10);
  ownerid = strtol (msg+1, const_cast <char **> (&msg), 10);
  fertility = pseudofloat (&msg);
  health = pseudofloat (&msg);
  age = strtod (msg+1, const_cast <char **> (&msg));
  fightstrenght = pseudofloat (&msg);
  targetx = strtod (msg+1, const_cast <char **> (&msg));
  targety = strtod (msg+1, const_cast <char **> (&msg));
  x = strtod (msg+1, const_cast <char **> (&msg));
  y = strtod (msg+1, const_cast <char **> (&msg));
  while (*(++msg) != ' ');	// skip water, it's useless
  grain = pseudofloat (&msg);
  grass = pseudofloat (&msg);
  carcass = pseudofloat (&msg);
  while (*(++msg) != ' ');	// skip hay, it's useless
  meat = pseudofloat (&msg);
  bread = pseudofloat (&msg);
  forest = pseudofloat (&msg);
}

inline void Communicate::write_serv (const char *const buf, const size_t &len)
  const
{
  while (write (fd, buf, len) == -1)
    {
      if (errno == EINTR)
	continue;
      perror ("hivemind");
      exit (1);
    }
}

Communicate::Communicate (int &argc, char **&argv)
  : control (this), turf (this)
{
  while (1)
    {
      int c;
      static struct option long_options[] = {
#ifdef HAVE_LIBPTHREAD
	{"concurrent", 0, NULL, 'c'},
#endif
	{0, 0, 0, 0}};
      c = getopt_long (argc, argv, "c", long_options, NULL);
      if (c == -1)
	break;
      switch (c)
	{
#ifdef HAVE_LIBPTHREAD
	case 'c':
	  control.is_concurrent ();
	  break;
#endif
	}
    }
  if (optind == argc)
    {
#ifdef HAVE_LIBPTHREAD
      cout << "Usage: hivemind hostname [--concurrent]\n";
#else
      cout << "Usage: hivemind hostname\n";
#endif
      exit (0);
    }      

  struct hostent *ent;
  ent = gethostbyname (argv[optind]);
  if (!ent)
    {
      cerr << "hivemind: lookup failed for name " << argv[optind] << endl;
      exit (1);
    }
  fd = socket (ent->h_addrtype, SOCK_STREAM, 0);
  cout << "hivemind: contacting host " << ent->h_name << endl;
#ifdef AF_INET6
  switch (ent->h_addrtype)
    {
    case AF_INET:
	{
#endif
	  struct sockaddr_in servaddr = {0};
	  servaddr.sin_family = AF_INET;
	  servaddr.sin_port = htons (CADAVER_PORT);
	  memcpy (&servaddr.sin_addr, ent->h_addr, ent->h_length);
	  if (connect (fd, (struct sockaddr *) &servaddr, sizeof (servaddr)))
	    {
	      perror ("hivemind");
	      exit (1);
	    }
#ifdef AF_INET6
	  break;
	}
    case AF_INET6:
	{
	  struct sockaddr_in6 servaddr = {0};
	  servaddr.sin6_family = AF_INET6;
	  servaddr.sin6_port = htons (CADAVER_PORT);
	  memcpy (&servaddr.sin6_addr, ent->h_addr, ent->h_length);
	  if (connect (fd, (struct sockaddr *) &servaddr, sizeof (servaddr)))
	    {
	      perror ("hivemind");
	      exit (1);
	    }
	}
    }
#endif
  cout << "hivemind: contact established\n";
}

void Communicate::wait_input_and_query (bool &read_finished)
{
  static Time sleep_time, next_tick_time;
  static bool start_tick = true;
  static int maxfd = fd+1;
  fd_set read_set;
  int retval;

  while (1)
    {
      if (start_tick)
	{
	  next_tick_time.now_time ();
	  next_tick_time += pulse_interval;
	  sleep_time = pulse_interval;
	  start_tick = false;
	  query ();
	}

      FD_ZERO (&read_set);
      FD_SET (fd, &read_set);
      retval = select (maxfd, &read_set, NULL, NULL, sleep_time);
      if (retval == 0)
	start_tick = true;
      else if (retval == 1 || errno == EINTR)
	{
	  Time reset_time;
	  reset_time.now_time ();
	  if (reset_time >= next_tick_time)
	    start_tick = true;
	  else
	    sleep_time = (next_tick_time - reset_time);
	  if (retval == 1)	// not when interrupted
	    return;
	}
      else
	{
	  perror ("hivemind");
	  control.~Control ();
	  exit (1);
	}
    }
}

void Communicate::readline (string &buf)
{
  static ssize_t buf_end, rawbufref;
  static char rawbuf[BUF_SIZE];

  while (1)
    {
      if (rawbufref == buf_end)
	{
	  buf_end = read (fd, rawbuf, BUF_SIZE);
	  if (buf_end == -1)
	    {
	      perror ("hivemind");
	      control.~Control ();
	      exit (1);
	    }
	  if (buf_end == 0)
	    {
	      cout << "hivemind: received EOF from server\n";
	      control.~Control ();
	      exit (0);		// not my fault
	    }
	  rawbufref = 0;
	}
      do
	{
	  buf += rawbuf[rawbufref];
	  if (rawbuf[rawbufref] == '\n' || rawbuf[rawbufref] == '\0')
	    {
	      buf += '\0';
	      rawbufref++;
	      return;
	    }
	} while (++rawbufref < buf_end);
    }
}

void Communicate::handler ()
{
  bool flip = 0;	// Fix this damn thing !!!
  bool read_finished = true;	// Nice try...
  string buf;
  turf.init (fd, id);		// Reads map size, returns client id
  // I'd really prefer another place for this
  add_component (new Nomad (this, &control, &turf));

#ifdef HAVE_LIBPTHREAD
  if (control.cc ())
    {
      for (vector<Component *>::iterator iter = componentlist.begin ();
	   iter != componentlist.end (); iter++)
	(*iter)->start_handler (/*static_cast<Nomad *>*/ (*iter));
      control.waitfor_default ();
    }
#endif
  while (1)
    {
      wait_input_and_query (read_finished);
      readline (buf);

      if (buf[0] == 'R')
	{
	  if (buf[1] == 'M')
	    turf.update (buf.c_str ()+2);
	  else if (buf[1] == 'C')
	    {
	      cyborg_update (buf.c_str ()+2);
#ifdef HAVE_LIBPTHREAD
	      if (!control.cc ())
#endif
	      if (flip)
		for (vector<Component *>::iterator iter =
		       componentlist.begin ();
		     iter != componentlist.end (); iter++)
		  (*iter)->handler ();
	      //read_finished = true;
	      flip = !flip;
	    }
	}
#if 0
      else if (buf[0] == 'M')
	cout << buf;		// It's nice to debug...
#endif
      buf.erase (buf.begin (), buf.end ());
    }
}

void Communicate::cyborg_update (const char *msg)
{
  set<int> missing = exists;

  while (*msg != '\n')
    {
      const Cyborg *cptr = new const Cyborg (msg);
      missing.erase (cptr->id);
      exists.insert (cptr->id);
      control.update_cyborg (cptr);	// Cyborg deletion handled also here
    }

  // Any cyborgs not listed are either dead or out of visibility range.
  // Either case, they need to be removed from the lists.
  for (set<int>::iterator iter = missing.begin (); iter != missing.end ();
       iter++)
    {
      control.cyborg_lost (*iter);
      exists.erase (*iter);
    }      
}

int main (int argc, char **argv)
{
  srand (time(NULL));
  Communicate *comm = new Communicate (argc, argv);
  comm->handler ();
}

void Communicate::take (const int &id, const enum Cyborg::item &item) const
{
  char buf[32];
  int len = snprintf (buf, 32, "TA %d %d\n", id, static_cast<int> (item));
  write_serv (buf, len);
}

void Communicate::drop (const int &id, const enum Cyborg::item &item) const
{
  char buf[32];
  int len = snprintf (buf, 32, "DR %d %d\n", id, static_cast<int> (item));
  write_serv (buf, len);
}

void Communicate::eat (const int &id, const enum Cyborg::item &item) const
{
  char buf[32];
  int len = snprintf (buf, 32, "EA %d %d\n", id, static_cast<int> (item));
  write_serv (buf, len);
}

void Communicate::convert (const int &id, const enum Cyborg::item &item) const
{
  char buf[32];
  int len = snprintf (buf, 32, "CO %d %d\n", id, static_cast<int> (item));
  write_serv (buf, len);
}

void Communicate::burn (const int &id, const enum Cyborg::item &item) const
{
  char buf[32];
  int len = snprintf (buf, 32, "BU %d %d\n", id, static_cast<int> (item));
  write_serv (buf, len);
}

void Communicate::idle (const int &id) const
{
  char buf[32];
  int len = snprintf (buf, 32, "MI %d\n", id);
  write_serv (buf, len);
}

void Communicate::move (const int &id, const float &x, const float &y) const
{
  char buf[32];
  int len = snprintf (buf, 32, "MG %d %.2f %.2f\n", id, x, y);
  write_serv (buf, len);
}

void Communicate::fight (const int &id, const int &enemyid) const
{
  char buf[32];
  int len = snprintf (buf, 32, "MF %d %d\n", id, enemyid);
  write_serv (buf, len);
}

void Communicate::sex (const int &id, const int &partnerid) const
{
  char buf[32];
  int len = snprintf (buf, 32, "MS %d %d\n", id, partnerid);
  write_serv (buf, len);
}
