/*  hivemind -- a client for cadaver
    Copyright (C) 2000 Kari Pahula

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
*/

#include <config.h>

#ifdef HAVE_LIBPTHREAD
#include <pthread.h>
#endif

#include <vector>
#include <algorithm>

#include "control.H"

#ifdef HAVE_LIBPTHREAD
RWlock::RWlock ()
{
  pthread_mutex_init (&nlock_mutex, NULL);
  pthread_cond_init (&w_wait, NULL);
  pthread_cond_init (&r_wait, NULL);
  write_lock = false;
  n_lock = 0;
}

void RWlock::get_read ()
{
  pthread_mutex_lock (&nlock_mutex);
  while (write_lock)
    pthread_cond_wait (&r_wait, &nlock_mutex);
  n_lock++;
  pthread_mutex_unlock (&nlock_mutex);
}

void RWlock::release_read ()
{
  pthread_mutex_lock (&nlock_mutex);
  if (--n_lock == 0 && write_lock)
    pthread_cond_signal (&w_wait);
  pthread_mutex_unlock (&nlock_mutex);
}

void RWlock::get_write ()
{
  pthread_mutex_lock (&nlock_mutex);
  write_lock = true;
  while (n_lock != 0)
    pthread_cond_wait (&w_wait, &nlock_mutex);
  pthread_mutex_unlock (&nlock_mutex);
}

void RWlock::release_write ()
{
  pthread_mutex_lock (&nlock_mutex);
  write_lock = false;
  pthread_cond_broadcast (&r_wait);
  pthread_mutex_unlock (&nlock_mutex);
}

#if 0
void Component::start_thread (void * (*startup)(void *), Component *data)
{
  pthread_attr_t attr;
  pthread_t thread;
  pthread_attr_init (&attr);
  pthread_attr_setdetachstate (&attr, PTHREAD_CREATE_DETACHED);
  pthread_create (&thread, &attr, startup, data);
}
#endif

template <class comp_type> static void * startup (void *this_ptr)
{
  static_cast<comp_type> (this_ptr)->handler ();
}

void Control::iam_default (const Component *me)
{
  if (cc ())
    {
      pthread_mutex_lock (&def_mutex);
      default_component = me;
      pthread_cond_signal (&def_cond);
      pthread_mutex_unlock (&def_mutex);
    }
  else
    default_component = me;
}

void Control::giveme_enemy (const Component *me) {}  // Make me!

void Control::waitfor_default ()
{
  if (cc ())
    {
      pthread_mutex_lock (&def_mutex);
      if (!default_component)
	pthread_cond_wait (&def_cond, &def_mutex);
      pthread_mutex_unlock (&def_mutex);
    }
}
#endif

void Control::refresh_list (Component *comp)
{
  /* First delete the cyborgs, then the entries.  */
  for (vector<const Cyborg *>::iterator iter = comp->clist.begin ();
       iter != comp->clist.end (); iter++)
    delete (*iter);
  for (vector<const Cyborg *>::iterator iter = comp->graveyard.begin ();
       iter != comp->graveyard.end (); iter++)
    delete (*iter);
  for (vector<const Cyborg *>::iterator iter = comp->elist.begin ();
       iter != comp->elist.end (); iter++)
    delete (*iter);
  comp->clist.erase (comp->clist.begin (), comp->clist.end ());
  comp->graveyard.erase (comp->graveyard.begin (), comp->graveyard.end ());
  comp->elist.erase (comp->elist.begin (), comp->elist.end ());
#ifdef HAVE_LIBPTHREAD
  if (cc ())
    list_lock.get_read ();
#endif
  for (list<struct master_t>::iterator iter = masterlist.begin ();
       iter != masterlist.end (); iter++)
    if (iter->comp == comp)
      {
	if (comm->same_owner (iter->cyborg->ownerid))
	  comp->clist.push_back (new const Cyborg(*(iter->cyborg)));
	else	// It's an enemy
	  comp->elist.push_back (new const Cyborg(*(iter->cyborg)));
      }

  for (list<struct master_t>::iterator iter = graveyard.begin ();
       iter != graveyard.end (); iter++)
    if (iter->comp == comp && comm->same_owner (iter->cyborg->ownerid))
      {
	comp->graveyard.push_back (iter->cyborg);
	iter->cyborg = NULL;	// Heathen are not allowed to the graveyard
      }
  /* Erase every entry of removed cyborgs.  */
  while (excavate_cyborg ());

#ifdef HAVE_LIBPTHREAD
  if (cc ())
    list_lock.release_read ();
#endif
}

void Control::update_cyborg (const Cyborg * const cyborg)
{
  list<struct master_t>::iterator iter;
  for (iter = masterlist.begin (); iter != masterlist.end (); iter++)
    if (iter->cyborg->id == cyborg->id)
      {
#ifdef HAVE_LIBPTHREAD
	if (cc ())
	  list_lock.get_write ();
#endif
	delete (iter->cyborg);
	iter->cyborg = cyborg;
#ifdef HAVE_LIBPTHREAD
	if (cc ())
	  list_lock.release_write ();
#endif
	return;
      }
  /* A new cyborg.  */
  struct master_t master;
  if (comm->same_owner (cyborg->id))
    master.comp = default_component;
  else
    master.comp = default_enemycomp;
  master.cyborg = cyborg;
#ifdef HAVE_LIBPTHREAD
  if (cc ())
    list_lock.get_write ();
#endif
  masterlist.push_back (master);
#ifdef HAVE_LIBPTHREAD
  if (cc ())
    list_lock.release_write ();
#endif
}

void Control::cyborg_lost (const int &id)
{
  /* The cyborg MUST be in the list, since this function is the only one which
     removes them from the list and the calling function never calls this
     function twice for the same cyborg.  */
  list<struct master_t>::iterator iter;
  for (iter = masterlist.begin (); iter != masterlist.end (); iter++)
    if (iter->cyborg->id == id)
      break;

#ifdef HAVE_LIBPTHREAD
  if (cc ())
    list_lock.get_write ();
#endif
  graveyard.push_back (*iter);
  masterlist.erase (iter);
#ifdef HAVE_LIBPTHREAD
  if (cc ())
    list_lock.release_write ();
#endif
}

const Cyborg *
Component::nearest_cyborg (const int &intx, const int &inty, float &dist,
			   void *thisptr,
			   bool (* const test) (void *, const Turf *const,
						const Cyborg *const, bool &))
{
  bool awayok;
  const Cyborg *cyborg = NULL;
  float x = intx+.5, y = inty+.5;
  float min = 10000000000.0;
  for (vector <const Cyborg *const>::iterator iter = clist.begin ();
       iter != clist.end (); iter++)
    if (test (thisptr, turf, *iter, awayok))
      {
	if (intx == static_cast<int> ((*iter)->x)
	    && inty == static_cast<int> ((*iter)->y))
	  {
	    dist = 0;
	    return *iter;
	  }
	else if (awayok)
	  {
	    float xval = x-(*iter)->x, yval = y-(*iter)->y;
	    dist = xval*xval+yval*yval;
	    if (dist < min)
	      {
		cyborg = *iter;
		min = dist;
	      }
	  }
      }
  dist = min;
  return cyborg;
}
