/*  hivemind -- a client for cadaver
    Copyright (C) 2000 Kari Pahula

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
*/

#include <config.h>

#define NOMAD_INTERVAL 200000

/* G == gather, W == weeds, H == harvest, S == seed  */

#define G_ACTIVE	(1<<0)
#define WG_ACTIVE	(1<<1)
#define WF_ACTIVE	(1<<2)
#define H_ACTIVE	(1<<3)
#define S_ACTIVE	(1<<4)
#define PERIMETER	(1<<5)
#define FARM		(1<<6)
#define NEW_FARM	(1<<7)

#include <unistd.h>
#ifdef HAVE_LIBPTHREAD
#include <pthread.h>
#endif

#include <map>
#include <vector>
#include <list>

#include "nomad.H"
#include "control.H"
#include "turf.H"
#include "communicate.H"

#if 0
void *Nomad::startup (void *ptr)
{
  while (1)
    {
      static_cast <Nomad *> (ptr)->handler ();
      usleep (NOMAD_INTERVAL);
    }
}
#endif

Nomad::Nomad (const Communicate *const co, Control *const ct,
	      const Turf *const tu)
{
  comm = co;
  control = ct;
  turf = tu;
  control->iam_default (this);
  control->giveme_enemy (this);
  map = new char [turf->size ()];
  for (int i = 0; i < turf->size (); i++)
    map[i] = 0;
  bread_stockpile_amount = 0;
}

inline void Nomad::abort_action (const Cyborg &cyborg)
{
  if (cyborg_act[cyborg.id])
    {
      int ref;
      if ((cyborg.state == Cyborg::GOTO
	   && (ref = turf->coord (cyborg.targetx, cyborg.targety)) != -1)
	  || (ref = turf->coord (cyborg.x, cyborg.y)) != -1)
	map[ref] &= ~cyborg_act[cyborg.id];
      cyborg_act[cyborg.id] = 0;
    }
}

/* This is the main heuristic routine of hivemind.  It basically first
   iterates through the list of cyborgs and initiates some important actions
   and halts finished actions.  On the second pass the routine iterates
   through every map cell, which has been designated to be a part of the
   'civilization'.  It dispatches idle cyborgs to handle any events.  The
   third and final pass iterates through the cyborgs again and this time
   assigns any cyborgs, which are still idle, to handle less urgent tasks.
*/
void Nomad::handler ()
{
  control->refresh_list (this);
  vector<const Cyborg *>::iterator iter;
  bool mates_avail = 1;

  /* Dig some graves.  */
  for (iter = graveyard.begin (); iter != graveyard.end (); iter++)
    {
      abort_action (**iter);
      cyborg_act.erase ((*iter)->id);
      cyborg_eat.erase ((*iter)->id);
    }

  /* First pass.  */
  for (iter = clist.begin (); iter != clist.end (); iter++)
    {
      nourisher (**iter);	// Don't starve!
      // If certain conditions are met, stop action
      if (action_stopper (**iter))
	{
	  (*iter)->state = Cyborg::IDLE;
	  comm->idle ((*iter)->id);	// Be extra sure
	}
      taker (**iter);		// Pick up any items here (harvest, weeding)
      marshal (**iter);		// It's a Quake out there
    }

  /* Second pass, keep the society running.  */
  planner ();

  /* Third pass, make every cyborg do something.  */
  for (iter = clist.begin (); iter != clist.end(); iter++)
    {
      if (!matchmaker (**iter, mates_avail)	// More cyborgs
	  && (*iter)->state == Cyborg::IDLE
	  && !farmjob (**iter))			// Then grow some crops
	find_resource (**iter);			// Then scavenge
    }
}

bool Nomad::test_any_food (const Turf &turf, void *compptr,
			   unsigned const int &x, unsigned const int &y)
{
  int ref = turf.coord (x, y);
  if (ref != -1)
    {
      const struct Turf::map_cell *cell = turf.read_map (ref);
      if (cell->meat>0 || cell->bread>0)
	return true;
    }
  return false;
}

void Nomad::nourisher (const Cyborg &cyborg)
{
  if (cyborg.state == Cyborg::EAT)
    {
      if (cyborg.health > 8)
	{
	  comm->idle (cyborg.id);	// Be extra sure
	  cyborg.state = Cyborg::IDLE;
	}
      return;
    }
  if (cyborg.health > 3 || cyborg.age > 1400)
    return;		// I'm OK or too old to bother.
  int ref = turf->coord (cyborg.x, cyborg.y);
  const struct Turf::map_cell *cell = turf->read_map (ref);
  // First let the cyborg gather an amount of food before doing anything
  if ((cyborg.state == Cyborg::TAKE || cyborg.state == Cyborg::CONVERT)
	   && cyborg.item == Cyborg::BREAD && cyborg.bread >= 4)
    comm->eat (cyborg.id, Cyborg::BREAD);
  else if (cyborg.state == Cyborg::TAKE && cyborg.item == Cyborg::MEAT
	   && cyborg.meat >= 2)
    comm->eat (cyborg.id, Cyborg::MEAT);
  else if (cyborg.meat)
    comm->eat (cyborg.id, Cyborg::MEAT);
  else if (cyborg.bread)
    comm->eat (cyborg.id, Cyborg::BREAD);
  else if (cell->bread)
    comm->take (cyborg.id, Cyborg::BREAD);
  else if (cell->meat)
    comm->take (cyborg.id, Cyborg::MEAT);
  else if (cyborg.grain)
    comm->convert (cyborg.id, Cyborg::BREAD);
  else		// We'll have to search further...
    {
      unsigned int x = static_cast<unsigned int> (cyborg.x);
      unsigned int y = static_cast<unsigned int> (cyborg.y);
      if (turf->find_nearest (x, y, 10, this, &test_any_food))
	comm->move (cyborg.id, x+.5, y+.5);
      else
	return;		// Just hope there's food available later...
    }
  abort_action (cyborg);	// Now I'm going to eat in any case!
  cyborg_eat.insert (cyborg.id);
  return;
}

void Nomad::taker (const Cyborg &cyborg)
{
  if (cyborg.state == Cyborg::GOTO)
    return;
  switch (cyborg_act[cyborg.id])
    {
    case H_ACTIVE:
      if (cyborg.state == Cyborg::TAKE && cyborg.item == Cyborg::GRAIN)
	break;
      comm->take (cyborg.id, Cyborg::GRAIN);
      cyborg.state = Cyborg::TAKE;
      break;
    case WF_ACTIVE:
      if (turf->read_map (cyborg.x, cyborg.y)->forest>0)
	{
	  if (cyborg.state == Cyborg::TAKE && cyborg.item == Cyborg::FOREST)
	    break;
	  comm->take (cyborg.id, Cyborg::FOREST);
	  cyborg.state = Cyborg::TAKE;
	}
      else
	cyborg_act[cyborg.id] = 0;
      break;
    case WG_ACTIVE:
      if (turf->read_map (cyborg.x, cyborg.y)->grass>0)
	{
	  if (cyborg.state == Cyborg::TAKE && cyborg.item == Cyborg::GRASS)
	    break;
	  comm->take (cyborg.id, Cyborg::GRASS);
	  cyborg.state = Cyborg::TAKE;
	}
      else
	cyborg_act[cyborg.id] = 0;
      break;
    case S_ACTIVE:
      if (turf->read_map (cyborg.x, cyborg.y)->grain <= 5 && cyborg.grain > 0)
	{
	  if (cyborg.state == Cyborg::DROP && cyborg.item == Cyborg::GRAIN)
	    break;
	  comm->drop (cyborg.id, Cyborg::GRAIN);
	  cyborg.state = Cyborg::DROP;
	}
      else
	cyborg_act[cyborg.id] = 0;	// action_stopper does the rest
      break;
    }
}

/* Ok, here's simple, stupid fighting...  */
void Nomad::marshal (const Cyborg &cyborg)
{
  if (cyborg_eat.count (cyborg.id))
    return;
  for (vector<const Cyborg *>::iterator iter = elist.begin ();
       iter != elist.end (); iter++)
    if (cyborg.dist (**iter) < 4.0)
      {
	abort_action (cyborg);	/* This'll potentially cause a flow of
				   cyborgs, as they come to continue the
				   aborted task.  */
	comm->fight (cyborg.id, (*iter)->id);
      }
}

bool Nomad::action_stopper (const Cyborg &cyborg)
{
  if (cyborg.state == Cyborg::IDLE)
    return false;	// Quick exit
  unsigned int ref = turf->coord (static_cast<const unsigned int> (cyborg.x),
				  static_cast<const unsigned int> (cyborg.y));
  const struct Turf::map_cell *cell = turf->read_map (ref);
  /* Don't be greedy.  */
  if (cyborg.item == Cyborg::BREAD
      && ((cyborg.state == Cyborg::CONVERT && cyborg.bread >= 4)
	  || (cyborg.state == Cyborg::DROP
	      && cell->bread >= bread_stockpile_amount)))
    return true;

  if (map[ref] & FARM && cyborg.item == Cyborg::GRAIN)
    {
      if (cyborg.state == Cyborg::TAKE)
	{
	  if (cell->grain < 7)	  /* Enough harvested, now let it grow.  */
	    {
	      map[ref] &= ~H_ACTIVE;
	      cyborg_act[cyborg.id] = 0;
	      return true;
	    }
	  else
	    return false;	// Finish the job.
	}
      else if (cyborg.state == Cyborg::DROP && cell->grain >= 6)
	{
	  map[ref] &= ~(NEW_FARM|S_ACTIVE);
	  return true;
	}
    }
  return false;
}

bool Nomad::ntest_idle (void *thisptr, const Turf *const turf,
			const Cyborg *const cyborg, bool &awayok)
{
  awayok = !(static_cast<Nomad *>
	     (thisptr)->map[turf->coord (cyborg->x, cyborg->y)] & NEW_FARM);
  return static_cast<Nomad *> (thisptr)->cyborg_act[cyborg->id] == 0
    && (cyborg->state == Cyborg::IDLE || cyborg->state == Cyborg::BURN
	|| (!static_cast<Nomad *> (thisptr)->cyborg_eat.count (cyborg->id)
	    && cyborg->state == Cyborg::CONVERT));
}

bool Nomad::ntest_idle_has_grain (void *thisptr, const Turf *const turf,
				  const Cyborg *const cyborg, bool &awayok)
{
  awayok = !(static_cast<Nomad *>
	     (thisptr)->map[turf->coord (cyborg->x, cyborg->y)] & NEW_FARM);
  return static_cast<Nomad *> (thisptr)->cyborg_act[cyborg->id] == 0
    && cyborg->grain > 0
    && (cyborg->state == Cyborg::IDLE || cyborg->state == Cyborg::BURN
	|| (!static_cast<Nomad *> (thisptr)->cyborg_eat.count (cyborg->id)
	    && cyborg->state == Cyborg::CONVERT));
}

void Nomad::gatherer (const unsigned int &ref, const enum Cyborg::item &item,
		      const bool &take, const unsigned char &mask,
	              bool &idle_bots,
		      bool (* const test) (void *, const Turf *const,
					   const Cyborg *const, bool &))
{
  if (map[ref] & mask)
    return;
  unsigned int x = turf->reftox (ref), y = turf->reftoy (ref);
  float dist;
  const Cyborg *cyborg = nearest_cyborg (x, y, dist, this, test);
  if (cyborg)
    {
      if (dist > 50.0)
	return;
      abort_action (*cyborg);
      if (x == static_cast<unsigned int> (cyborg->x)
	  && y == static_cast<unsigned int> (cyborg->y))
	{
	  if (take)
	    comm->take (cyborg->id, item);
	  else
	    comm->drop (cyborg->id, item);
	  cyborg->state = Cyborg::TAKE;
	}
      else
	{
	  comm->move (cyborg->id, x+.5, y+.5);
	  cyborg->state = Cyborg::GOTO;
	}
      cyborg_act[cyborg->id] = mask;
      map[ref] |= mask;
    }
  else
    idle_bots = false;
}

void Nomad::planner ()
{
  bool idle_bots = true;
 restart:
  for (vector<unsigned int>::iterator iter = civilization.begin ();
       iter != civilization.end (); iter++)
    {
      const struct Turf::map_cell *cell = turf->read_map (*iter);
      /* This happens so rarely, that the vector container type is feasible.
	 If a cyborg has abandoned a farm for any reason, and it is overrun by
	 weeds, it is no longer considered a farm.  This could also happen if
	 the map is randomized.  */
      if (!(map[*iter] & NEW_FARM) && (cell->grass >= 5 || cell->forest >= 5))
	{
	  map[*iter] &= G_ACTIVE;	// Erase other flags
	  map[*iter] |= PERIMETER;	// Probably near other farms
	  if (iter == civilization.begin ())
	    {
	      civilization.erase (iter);	// End of the beginning
	      goto restart;
	    }
	  else
	    {
	      vector<unsigned int>::iterator erased = iter--;
	      civilization.erase (erased);
	      continue;
	    }
	}
      if (idle_bots)
	{
#if 0
	  if ((((map[*iter] & NEW_FARM)
		&& cell->grass <= 2 && cell->forest <= 2)
	       || !(map[*iter] & NEW_FARM))
	      && cell->grain <= 5)
#endif
	  if (cell->grain <= 5 && !cell->grass && !cell->forest)
	    {
	      if (!(map[*iter] & S_ACTIVE))
		gatherer (*iter, Cyborg::GRAIN, false, S_ACTIVE, idle_bots,
			  &ntest_idle_has_grain);
	    }
	  else
	    map[*iter] &= ~S_ACTIVE;
	  if (!idle_bots)
	    continue;
	  if (cell->grain > 8 && !(map[*iter] & H_ACTIVE))
	    gatherer (*iter, Cyborg::GRAIN, true, H_ACTIVE, idle_bots,
		      &ntest_idle);
	  if (!idle_bots)
	    continue;
	  /* Weeding action unmarking is done here, since a cyborg finished
	     with weeding becomes automatically idle anyway.  */
	  if (cell->forest)
	    {
	      if (!(map[*iter] & WF_ACTIVE))
		gatherer (*iter, Cyborg::FOREST, true, WF_ACTIVE, idle_bots,
			  &ntest_idle);
	    }
	  else
	    map[*iter] &= ~WF_ACTIVE;
	  if (!idle_bots)
	    continue;
	  if (cell->grass)
	    {
	      if (!(map[*iter] & WG_ACTIVE))
		gatherer (*iter, Cyborg::GRASS, true, WG_ACTIVE, idle_bots,
			  &ntest_idle);
	    }
	  else
	    map[*iter] &= ~WG_ACTIVE;
	}
    }

  if (!civilization.empty ())
    {
      bread_stockpile_amount = (16*clist.size ())/(civilization.size ());
      if (bread_stockpile_amount > 9)
	bread_stockpile_amount = 9;
    }
}

bool Nomad::matchmaker (const Cyborg &cyborg, bool &mates_avail)
{
  if (mates_avail && !cyborg.gender && cyborg.fertility == 0)
    {
      vector<const Cyborg *>::iterator candidate;
      unsigned int x = static_cast<unsigned int> (cyborg.x);
      unsigned int y = static_cast<unsigned int> (cyborg.x);
      float min_dist = 1<<30, new_dist;		// make min_dist big
      vector<const Cyborg *>::iterator mate = NULL;
      for (candidate = clist.begin (); candidate != clist.end ();
	   candidate++)
	if ((*candidate)->gender && (*candidate)->fertility == 0
	    && (new_dist=cyborg.dist (**candidate)) < min_dist)
	  {
	    mate = candidate;
	    min_dist = new_dist;
	  }
      if (!mate)
	{
	  mates_avail = 0;	// No mates available right now for anybody
	  return false;
	}
      else if (min_dist < 4.0)	// 2*2
	{
	  comm->sex ((*mate)->id, cyborg.id);
	  return false;		// get on with normal life
	}
#if 0		/* Disabled for now...  */
      /* Farmers are a more complex case...  */
      else if (!(map[turf->coord (cyborg.x, cyborg.y)] & FARM)
	       || !turf->find_nearest (x, y, 4, this, &test_farm))
	{
	  /* Move only the female, to disturb other actions as little as
	     possible.  */
	  comm->move (cyborg, (*mate)->x, (*mate)->y);
	  return true;
	}
#endif
    }
  return false;
}

bool Nomad::test_perimeter (const Turf &turf, void *compptr,
			   unsigned const int &x, unsigned const int &y)
{
  int ref = turf.coord (x, y);
  if (ref != -1 && (static_cast<Nomad *> (compptr)->map[ref] & PERIMETER))
    return true;
  return false;
}

bool Nomad::test_can_make_field (const Turf &turf, void *compptr,
				 unsigned const int &x, unsigned const int &y)
{
  int ref = turf.coord (x, y);
  if (ref != -1 && !(static_cast<Nomad *> (compptr)->map[ref] & FARM))
    return true;
  return false;
}

/* These two tests determine, if a farm can hold more food.  */
bool Nomad::test_farm_more_bread (const Turf &turf, void *compptr,
				  unsigned const int &x, unsigned const int &y)
{
  int ref = turf.coord (x, y);
  if (ref != -1 && (static_cast<Nomad *> (compptr)->map[ref] & FARM)
      && (turf.read_map (ref)->bread <
	  static_cast<Nomad *> (compptr)->bread_stockpile_amount))
    return true;
  return false;
}

bool Nomad::test_farm_more_meat (const Turf &turf, void *compptr,
				 unsigned const int &x, unsigned const int &y)
{
  int ref = turf.coord (x, y);
  if (ref != -1 && (static_cast<Nomad *> (compptr)->map[ref] & FARM)
      && turf.read_map (ref)->meat < 9)
    return true;
  return false;
}

bool Nomad::farmjob (const Cyborg &cyborg)
{
  unsigned int x = static_cast <const unsigned int> (cyborg.x);
  unsigned int y = static_cast <const unsigned int> (cyborg.y);
  unsigned int ref = turf->coord (x, y);
  const struct Turf::map_cell *cell = turf->read_map (ref);
  cyborg_eat.erase (cyborg.id);		// If we're here, we're not eating

#if 0
  if ((map[ref] & FARM) && !(!(map[ref] & NEW_FARM)
			     || (cell->forest <= 2 && cell->grass <= 2))
      && cell->grain < 6 && cyborg.grain > 0)
    {
      comm->drop (cyborg.id, Cyborg::GRAIN);
      return true;
    }
#endif

  /* Handle new farms.  */
  if (cyborg.grain > 7 && !(map[ref] & NEW_FARM)
      && (clist.size ()*4) > civilization.size ())
    {
      if (!(map[ref] & PERIMETER))
	{	/* Females are supposed to start farms.  */
	  if (!turf->find_nearest (x, y, cyborg.gender ? 10:7, this,
				   &test_perimeter))
	    {		// Gather some more
	      if (cyborg.grain < (cyborg.gender ? 30:20))
		return false;
	      x = static_cast<unsigned const int> (cyborg.x);
	      y = static_cast<unsigned const int> (cyborg.y);
	      if (map[ref] & FARM
		  && turf->find_nearest (x, y, 7, this,
					 &test_can_make_field))
		{
		  comm->move (cyborg.id, x+.5, y+.5);
		  return true;
		}
	    }
	  else
	    {
	      comm->move (cyborg.id, x+.5, y+.5);
	      return true;
	    }
	}
      make_farm (ref, cyborg);
      return true;
    }
  x = static_cast<unsigned const int> (cyborg.x);
  y = static_cast<unsigned const int> (cyborg.y);
  /* Handle all the other chores of a farmer.  */
  if ((map[ref] & FARM) || turf->find_nearest (x, y, 4, this, &test_farm))
    {
      if (cyborg.bread < 4 && cyborg.grain)
	{	/* Bake some bread.  */
	  comm->convert (cyborg.id, Cyborg::BREAD);
	  return true;
	}
      else if (cyborg.carcass)
	{	/* Mmm... meatloaf...  */
	  comm->convert (cyborg.id, Cyborg::MEAT);
	  return true;
	}
      else if (cyborg.forest)
	{
	  comm->burn (cyborg.id, Cyborg::FOREST);
	  return true;
	}
      else if (cyborg.grass)
	{
	  comm->burn (cyborg.id, Cyborg::GRASS);
	  return true;
	}	/* Otherwise, put some food on farmland, where others can find
		   it.  */
      else if (cyborg.bread >= 4)
	{
	  if (cyborg.state == Cyborg::CONVERT
	      && cyborg.item == Cyborg::BREAD)
	    comm->idle (cyborg.id);
	  x = static_cast<unsigned const int> (cyborg.x);
	  y = static_cast<unsigned const int> (cyborg.y);
	  if (test_farm_more_bread (*turf, this, x, y))
	    {
	      comm->drop (cyborg.id, Cyborg::BREAD);
	      return true;
	    }
	  else if (turf->find_nearest (x, y, 7, this,
				       &test_farm_more_bread))
	    {
	      comm->move (cyborg.id, x+.5, y+.5);
	      return true;
	    }
	}
      else if (cyborg.meat >= 1)
	{
	  x = static_cast<unsigned const int> (cyborg.x);
	  y = static_cast<unsigned const int> (cyborg.y);
	  if (test_farm_more_meat (*turf, this, x, y))
	    {
	      comm->drop (cyborg.id, Cyborg::MEAT);
	      return true;
	    }
	  else if (turf->find_nearest (x, y, 7, this,
				       &test_farm_more_meat))
	    {
	      comm->move (cyborg.id, x+.5, y+.5);
	      return true;
	    }
	}
    }
  return false;
}

void Nomad::make_farm (const unsigned int &ref, const Cyborg &cyborg)
{
  unsigned int x = turf->reftox (ref), y = turf->reftoy (ref);
  int perimref;
  
  map[ref] = FARM|NEW_FARM;
  perimref = turf->coord (x+1, y);
  if (perimref != -1 && !(map[perimref] & FARM))
    map[perimref] |= PERIMETER;
  perimref = turf->coord (x-1, y);
  if (perimref != -1 && !(map[perimref] & FARM))
    map[perimref] |= PERIMETER;
  perimref = turf->coord (x, y+1);
  if (perimref != -1 && !(map[perimref] & FARM))
    map[perimref] |= PERIMETER;
  perimref = turf->coord (x, y-1);
  if (perimref != -1 && !(map[perimref] & FARM))
    map[perimref] |= PERIMETER;
  civilization.push_back (ref);
}

bool Nomad::test_farm (const Turf &turf, void *compptr,
		       unsigned const int &x, unsigned const int &y)
{
  unsigned int ref = turf.coord (x, y);
  if (ref != -1 && (static_cast<Nomad *> (compptr)->map[ref] & FARM))
    return true;
  return false;
}

bool Nomad::test_carcass (const Turf &turf, void *compptr,
			  unsigned const int &x, unsigned const int &y)
{
  unsigned int ref = turf.coord (x, y);
  if (ref != -1)
    {
      const struct Turf::map_cell *ptr = turf.read_map (ref);
      if (ptr->carcass > 0
	  && !(static_cast<Nomad *> (compptr)->map[ref] & G_ACTIVE))
	{
	  static_cast<Nomad *> (compptr)->map[ref] |= G_ACTIVE;
	  return true;
	}
    }
  return false;
}

bool Nomad::test_grain (const Turf &turf, void *compptr,
			unsigned const int &x, unsigned const int &y)
{
  unsigned int ref = turf.coord (x, y);
  if (ref != -1)
    {
      const struct Turf::map_cell *ptr = turf.read_map (ref);
      if (ptr->grain > 0
	  && !(static_cast<Nomad *> (compptr)->map[ref] & (FARM|G_ACTIVE)))
	{
	  static_cast<Nomad *> (compptr)->map[ref] |= G_ACTIVE;
	  return true;
	}
    }
  return false;
}

bool Nomad::test_food (const Turf &turf, void *compptr,
		       unsigned const int &x, unsigned const int &y)
{
  unsigned int ref = turf.coord (x, y);
  if (ref != -1)
    {
      const struct Turf::map_cell *ptr = turf.read_map (ref);
      if ((ptr->meat > 0 || ptr->bread > 0)
	  && !(static_cast<Nomad *> (compptr)->map[ref] & (FARM|G_ACTIVE)))
	{
	  static_cast<Nomad *> (compptr)->map[ref] |= G_ACTIVE;
	  return true;
	}
    }
  return false;
}

void Nomad::find_resource (const Cyborg &cyborg)
{
  unsigned int range, farm_x, farm_y;
  bool near_farm = false;
  static bool (*funcs[3]) (const Turf &, void *, unsigned const int &,
			   unsigned const int &) =
  {&test_carcass, &test_grain, &test_food};
  unsigned int x = static_cast<unsigned int> (cyborg.x);
  unsigned int y = static_cast<unsigned int> (cyborg.y);
  int ref = turf->coord (x, y);
  range = 5 - turf->find_nearest (x, y, 5, this, &test_farm);
  if (range == 5)
    range = 7;
  else
    {
      farm_x = x;
      farm_y = y;
      near_farm = true;
    }

  for (int i = 0; i < 3; i++)	// Search in order: carcass, grain, any food
    {
      x = static_cast<unsigned int> (cyborg.x);
      y = static_cast<unsigned int> (cyborg.y);
      funcs[i] (*turf, this, x, y);	// Mark the cell we're on
      if (map[ref] & G_ACTIVE)
	{
	  const struct Turf::map_cell *cell = turf->read_map (ref);
	  if (cell->carcass > 0)
	    comm->take (cyborg.id, Cyborg::CARCASS);
	  else if (cell->grain > 0)
	    comm->take (cyborg.id, Cyborg::GRAIN);
	  else if (cell->meat > 0)
	    comm->take (cyborg.id, Cyborg::MEAT);
	  else if (cell->bread > 0)
	    comm->take (cyborg.id, Cyborg::BREAD);
	  else
	    map[ref] &= ~G_ACTIVE;	// No more resources.
	  return;
	}
      else if (range > 0 && turf->find_nearest (x, y, range, this, funcs[i]))
	{
	  comm->move (cyborg.id, x+.5, y+.5);
	  return;
	}
    }
  if (map[ref] < FARM && near_farm)
    comm->move (cyborg.id, farm_x+.5, farm_y+.5);
}
